#!/usr/bin/env python

from selenium.webdriver.common.action_chains import ActionChains
from selenium.webdriver.common.actions.pointer_input import PointerInput
from selenium import webdriver
from selenium.webdriver.remote.webelement import WebElement
from selenium_firefox import FirefoxBinary, FirefoxLogInterceptor, Options
from HLISA.hlisa_action_chains import HLISA_ActionChains
import os.path
import sys
import time

fb = FirefoxBinary(firefox_path="firefox/firefox")
webdriver = webdriver.Firefox(firefox_binary=fb)


webdriver.maximize_window()
webdriver.set_page_load_timeout(30)


webdriver.get('http://localhost:8000/')

# Clicking test:
actions = HLISA_ActionChains(webdriver)
actions.scroll_by(0, 100)
button1 = webdriver.find_element_by_id("button1")
for i in range(10):
    actions.click(button1)

# Context clicking test:

button2 = webdriver.find_element_by_id("button2")
actions.context_click(button2)

# Test whether cursor location is correct after scroll:

actions.scroll_by(0, 100)
button3 = webdriver.find_element_by_id("button3")
actions.click(button3)
        
actions.perform()

# Test whether the cursor location is correcter after a different page has been loaded:

webdriver.get('http://localhost:8000/page2.html')

actions.reset_actions()
button1 = webdriver.find_element_by_id("button1")
actions.click(button1)

# Test drag and drop:

draggable_element = webdriver.find_element_by_id("draggable_element")
drag_endpoint = webdriver.find_element_by_id("drag_endpoint")
actions.drag_and_drop(draggable_element, drag_endpoint)

# Element with offset test:

element_with_offset = webdriver.find_element_by_id("element_with_offset")
actions.move_to_element_with_offset(element_with_offset, 12, 34)
actions.click()

# Double click test:

double_click_element = webdriver.find_element_by_id("double_click")
actions.double_click(double_click_element)

# Test drag and drop with offset:

draggable_element_offset = webdriver.find_element_by_id("draggable_element_offset")
drag_endpoint_offset = webdriver.find_element_by_id("drag_endpoint_offset")
actions.drag_and_drop_by_offset(draggable_element_offset, 500, 80)

# Send keys to element test:

send_keys_to_element_test = webdriver.find_element_by_id("send_keys_to_element_test")
actions.send_keys_to_element(send_keys_to_element_test, "test test")

# Key up and key down test:

key_up_key_down_test = webdriver.find_element_by_id("key_up_key_down_test")
actions.key_down("a", key_up_key_down_test)
actions.key_up("a")
actions.key_down("b")
actions.key_up("b")

# End of tests

actions.pause(50)
actions.perform()


import math
import time
import random
import numpy as np
from HLISA.errors import (ElementBoundariesWereZeroException,
                            IllegalArgumentException,
                            NoCursorCoordinatesException)

from selenium.webdriver import Firefox
from selenium.webdriver.common.by import By
from selenium.webdriver.common.action_chains import ActionChains
from selenium.webdriver.common.actions.pointer_input import PointerInput
from selenium.webdriver.remote.webelement import WebElement

from selenium.common.exceptions import MoveTargetOutOfBoundsException

class HL_Util:
    def create_pointer_move(self, duration=50, x=None, y=None, origin=None):
        """ The function to replace the original Selenium function that does not support specifying the duration.
        """
        action = dict(type="pointerMove", duration=duration)
        action["x"] = x
        action["y"] = y
        if isinstance(origin, WebElement):
            action["origin"] = {"element-6066-11e4-a52e-4f735466cecf": origin.id}
        elif origin is not None:
            action["origin"] = origin

        self.add_action(action)

def behavorial_element_coordinates(webdriver, element):
    """ Takes an element and returns coordinates somewhere in the element. If the element is not visable, it returns None.
        Uses a normal distribution, to bias the click towards the center of the element.
    """
    x_relative = int(element.rect['x']) - get_current_scrolling_position(webdriver)["x"]
    y_relative = int(element.rect['y']) - get_current_scrolling_position(webdriver)["y"]
    viewport_width = webdriver.execute_script("return window.innerWidth")
    viewport_height = webdriver.execute_script("return window.innerHeight")
    if element.rect['width'] == 0 or element.rect['height'] == 0:
        error_msg = """
            The element's plane was zero. To avoid this issue, you may want to use:
                from HLISA.util import best_effort_element_selection as best_selection
                from HLISA.errors import ElementBoundariesWereZeroException
                try:
                    click_with_HLISA(driver, el)
                except ElementBoundariesWereZeroException:
                    best_el = best_selection(driver, el)
                    click_with_HLISA(driver, best_el[0])
        """
        raise ElementBoundariesWereZeroException(error_msg)
    for _ in range(100): # Try 10 random positions, as some positions are not in round buttons.
        x = x_relative + int(np.random.normal(int(element.rect['width']*0.5), int(element.rect['width']*0.2)))
        y = y_relative + int(np.random.normal(int(element.rect['height']*0.5), int(element.rect['height']*0.2)))
        
        if x < 0 or y < 0 or x > viewport_width or y > viewport_height:
            continue
        
        coords_in_button = webdriver.execute_script(f"return document.elementFromPoint({x}, {y}) === arguments[0];", element)
        coords_in_descendant = webdriver.execute_script(f"""
            let el = document.elementFromPoint({x}, {y});
            return [...arguments[0].querySelectorAll('*')].includes(el);""", element)

        if coords_in_button or coords_in_descendant:
            return (x, y)
    return None

def std_positive(mean, std, minimal):
    """ Returns a number from a normal distribution that is larger or equal to parameter 'minimal'.
        Due to the minimum value, the returned values will not form a normal distribution.
        To minimize this effect, values that would have been smaller than the minimum are not drawn
        again, but get added a random small value. The new number will never become larger as the mean.
    """
    if minimal > mean:
        raise IllegalArgumentException(
            f"\n\tMethod util.std_positive expects argument 'minimal' <= 'mean'\n" + \
            f"\tGot:\n\t\tminimal: {minimal}\n\t\tmean: {mean}"
        )
    sample = np.random.normal(mean, std)
    if sample < minimal:
        sample += (minimal-sample) + (mean - minimal)*random.random()
    return sample

def scale_delay_kwargs(scaling, delay_kwargs):
    return {
        'mean': delay_kwargs['mean'] * (1/scaling),
        'std': delay_kwargs['mean'] * (1/scaling),
        'minimal': delay_kwargs['mean'] * (1/scaling),
    }
    

def increaseMousemovementSpeed():
    """ Replace a function in the original Selenium API to increase mouse movement speed.
    """
    PointerInput.create_pointer_move = HL_Util.create_pointer_move

def get_current_scrolling_position(webdriver, element=None):
    """ Returns the x and y offset for scrolling
    """
    if element:
        return ({"x": int(element.get_attribute("scrollLeft")), "y": int(element.get_attribute("scrollTop"))})
    return webdriver.execute_script("return {'x':window.pageXOffset, 'y':window.pageYOffset};")

def get_scrollable_elements(webdriver, element):
    """
    """
    script = """
        let scrollNodes = [];
        let node = arguments[0];

        function getScrollParent(node) {
          if (node == null) {
            return null;
          }
          if (node.scrollHeight > node.clientHeight) {
              return node;
          } else {
            return getScrollParent(node.parentNode);
          }
        }

        while (node != null){
          node = getScrollParent(node.parentNode);
          if (node != null){
              scrollNodes.push(node);
          }
        }
        return scrollNodes;
    """
    return webdriver.execute_script(script, element)

def element_is_scrollable(webdriver, element):
    """
    """
    script = """
        let node = arguments[0];
        if (node == null) {
            return false;
        }
        return node.scrollHeight > node.clientHeight;
    """
    return webdriver.execute_script(script, element)

def best_effort_element_selection(webdriver, element):
    """ Collection of best efforts approaches to select a clickable element
    """
    if element.rect["height"] == 0 or element.rect["width"] == 0:
        script ="""
            let res = [];
            arguments[0].querySelectorAll('*').forEach(el => {
                if (el.clientWidth > 0 && el.clientHeight > 0){
                    res.push(el);
                }
            });
            return res;"""
        candidates = webdriver.execute_script(script, element)
        return candidates

# Use the coordinate information from the Selenium 
# MoveTargetOutOfBoundsException to get the coordinates
# of the Selenium mouse cursor.
def get_cursor_coordinates(driver):
    MOVE_PIXELS = 30000
    try:
        ac = ActionChains(driver)
        ac.move_by_offset(MOVE_PIXELS, MOVE_PIXELS)
        ac.perform()
    except MoveTargetOutOfBoundsException as ex:
        error_message = str(ex)[10:23]
        x = error_message[0:5]
        y = error_message[7:12]
        try:
            x = int(x)
            y = int(y)
        except Exception as e:
            raise NoCursorCoordinatesException() # If the coordinates are not integers, something went wrong
        if x < 0 or y < 0:
            raise NoCursorCoordinatesException()
        return (x - MOVE_PIXELS, y - MOVE_PIXELS)
    raise NoCursorCoordinatesException() # If no exception occured, something went wrong



'''
    HLISA Time (Delay) Constants

    When delays are in sets of 4 with the postfixes 
    `_MEAN`, `_STD`, `_MIN` & `_KWARGS they are used as
    argumenst to culculate a delay using HLISA.util.std_positive.
    The `_KWARGS` postfix is used for unpacking as keyword arguments.

'''



'''
    HL_Selenium_Actions.addDelayAfterAction delay constants.
'''
ACTION_DELAY_MEAN = 0.3
ACTION_DELAY_STD  = 0.1
ACTION_DELAY_MIN  = 0.025
ACTION_DELAY_KWARGS = {
    'mean': ACTION_DELAY_MEAN,
    'std': ACTION_DELAY_STD,
    'minimal': ACTION_DELAY_MIN
}

'''
    HL_Selenium_Actions.click delay for how long the mouse is held down
    during a normal click.
'''
CLICK_HOLD_DELAY_MEAN = 0.092
CLICK_HOLD_DELAY_STD  = 0.018
CLICK_HOLD_DELAY_MIN  = 0.0
CLICK_HOLD_DELAY_KWARGS = {
    'mean': CLICK_HOLD_DELAY_MEAN,
    'std': CLICK_HOLD_DELAY_STD,
    'minimal': CLICK_HOLD_DELAY_MIN
}

'''
    HL_Selenium_Actions.send_keys delay for how long to wait before adding the '.'
    at the end of a sentence.
'''
SENTENCE_CLOSING_DELAY_MEAN = 1.7
SENTENCE_CLOSING_DELAY_STD  = 0.7
SENTENCE_CLOSING_DELAY_MIN  = 0.3
SENTENCE_CLOSING_DELAY_KWARGS = {
    'mean': SENTENCE_CLOSING_DELAY_MEAN,
    'std': SENTENCE_CLOSING_DELAY_STD,
    'minimal': SENTENCE_CLOSING_DELAY_MIN
}

'''
    HL_Selenium_Actions.send_keys delay for how long to wait after completing
    an entire sentence including.
'''
SENTENCE_COMPLETION_DELAY_MEAN = 0.6
SENTENCE_COMPLETION_DELAY_STD  = 0.4
SENTENCE_COMPLETION_DELAY_MIN  = 0.05
SENTENCE_COMPLETION_DELAY_KWARGS = {
    'mean': SENTENCE_COMPLETION_DELAY_MEAN,
    'std': SENTENCE_COMPLETION_DELAY_STD,
    'minimal': SENTENCE_COMPLETION_DELAY_MIN
}



'''
    HL_Selenium_Actions.write_sentence delay for how long to wait starting a sentence.
'''
SENTENCE_OPENING_DELAY_MEAN = 1.3
SENTENCE_OPENING_DELAY_STD  = 1.0
SENTENCE_OPENING_DELAY_MIN  = 0.2
SENTENCE_OPENING_DELAY_KWARGS = {
    'mean': SENTENCE_OPENING_DELAY_MEAN,
    'std': SENTENCE_OPENING_DELAY_STD,
    'minimal': SENTENCE_OPENING_DELAY_MIN
}


'''
    HL_Selenium_Actions.write_sentence delay between characters
'''
SENTENCE_CHARACTER_DELAY_MEAN = 0.21
SENTENCE_CHARACTER_DELAY_STD  = 0.03
SENTENCE_CHARACTER_DELAY_MIN  = 0.011
SENTENCE_CHARACTER_DELAY_KWARGS = {
    'mean': SENTENCE_CHARACTER_DELAY_MEAN,
    'std': SENTENCE_CHARACTER_DELAY_STD,
    'minimal': SENTENCE_CHARACTER_DELAY_MIN
}


'''
    HL_Selenium_Actions.write_word delay for how long to wait starting a word.
'''
WORD_OPENING_DELAY_MEAN = 0.47
WORD_OPENING_DELAY_STD  = 0.21
WORD_OPENING_DELAY_MIN  = 0.05
WORD_OPENING_DELAY_KWARGS = {
    'mean': WORD_OPENING_DELAY_MEAN,
    'std': WORD_OPENING_DELAY_STD,
    'minimal': WORD_OPENING_DELAY_MIN
}

'''
    HL_Selenium_Actions.write_word delay for how long to wait after completing a word.
'''
WORD_CLOSING_DELAY_MEAN = 0.2
WORD_CLOSING_DELAY_STD  = 0.08
WORD_CLOSING_DELAY_MIN  = 0.01
WORD_CLOSING_DELAY_KWARGS = {
    'mean': WORD_CLOSING_DELAY_MEAN,
    'std': WORD_CLOSING_DELAY_STD,
    'minimal': WORD_CLOSING_DELAY_MIN
}

'''
    HL_Selenium_Actions.write_word delay between the characters in a word.
'''
WORD_CHARACTER_DELAY_MEAN = 0.21
WORD_CHARACTER_DELAY_STD  = 0.03
WORD_CHARACTER_DELAY_MIN  = 0.011
WORD_CHARACTER_DELAY_KWARGS = {
    'mean': WORD_CHARACTER_DELAY_MEAN,
    'std': WORD_CHARACTER_DELAY_STD,
    'minimal': WORD_CHARACTER_DELAY_MIN
}

'''
    HL_Selenium_Actions.write_character delay between the characters in a word.
'''
CHARACTER_SHIFT_DOWN_DELAY_MEAN = 0.06
CHARACTER_SHIFT_DOWN_DELAY_STD  = 0.035
CHARACTER_SHIFT_DOWN_DELAY_MIN  = 0.005
CHARACTER_SHIFT_DOWN_DELAY_KWARGS = {
    'mean': CHARACTER_SHIFT_DOWN_DELAY_MEAN,
    'std': CHARACTER_SHIFT_DOWN_DELAY_STD,
    'minimal': CHARACTER_SHIFT_DOWN_DELAY_MIN
}

'''
    HL_Selenium_Actions.write_character delay between the characters in a word.
'''
CHARACTER_SHIFT_UP_DELAY_MEAN = 0.03
CHARACTER_SHIFT_UP_DELAY_STD  = 0.015
CHARACTER_SHIFT_UP_DELAY_MIN  = 0.003
CHARACTER_SHIFT_UP_DELAY_KWARGS = {
    'mean': CHARACTER_SHIFT_UP_DELAY_MEAN,
    'std': CHARACTER_SHIFT_UP_DELAY_STD,
    'minimal': CHARACTER_SHIFT_UP_DELAY_MIN
}

'''
    HL_Selenium_Actions.write_character delay between the characters in a word.
'''
CHARACTER_DWELL_DELAY_MEAN = 0.06
CHARACTER_DWELL_DELAY_STD  = 0.035
CHARACTER_DWELL_DELAY_MIN  = 0.013
CHARACTER_DWELL_DELAY_KWARGS = {
    'mean': CHARACTER_DWELL_DELAY_MEAN,
    'std': CHARACTER_DWELL_DELAY_STD,
    'minimal': CHARACTER_DWELL_DELAY_MIN
}

'''
    HL_Selenium_Actions.key_down delay after the action if the addDelayAfter=True
'''
KEY_DOWN_DELAY_MEAN = 0.06
KEY_DOWN_DELAY_STD  = 0.035
KEY_DOWN_DELAY_MIN  = 0.013
KEY_DOWN_DELAY_KWARGS = {
    'mean': KEY_DOWN_DELAY_MEAN,
    'std': KEY_DOWN_DELAY_STD,
    'minimal': KEY_DOWN_DELAY_MIN
}

'''
    HL_Selenium_Actions.key_up delay after the action if the addDelayAfter=True
'''
KEY_UP_DELAY_MEAN = 0.21
KEY_UP_DELAY_STD  = 0.03
KEY_UP_DELAY_MIN  = 0.011
KEY_UP_DELAY_KWARGS = {
    'mean': KEY_UP_DELAY_MEAN,
    'std': KEY_UP_DELAY_STD,
    'minimal': KEY_UP_DELAY_MIN
}

'''
    The following delays are used with the numpy.random.normal function, and
    thus the '_KWARGS' post fix version have different keys that correspond 
    to the numpy implementation.
'''

'''
    HL_Selenium_Actions.double_click delay to hold the first click.
'''
DOUBLE_CLICK_FIRST_HOLD_DELAY_MEAN = 0.083
DOUBLE_CLICK_FIRST_HOLD_DELAY_STD  = 0.034
DOUBLE_CLICK_SECOND_HOLD_DELAY_KWARGS = {
    'loc': DOUBLE_CLICK_FIRST_HOLD_DELAY_MEAN,
    'scale': DOUBLE_CLICK_FIRST_HOLD_DELAY_STD 
}

'''
    HL_Selenium_Actions.double_click delay between the two clicks.
'''
DOUBLE_CLICK_BETWEEN_DELAY_MEAN = 0.075
DOUBLE_CLICK_BETWEEN_DELAY_STD  = 0.026
DOUBLE_CLICK_BETWEEN_DELAY_KWARGS = {
    'loc': DOUBLE_CLICK_BETWEEN_DELAY_MEAN,
    'scale': DOUBLE_CLICK_BETWEEN_DELAY_STD 
}


'''
    HL_Selenium_Actions.double_click delay to hold the second click.
'''
DOUBLE_CLICK_SECOND_HOLD_DELAY_MEAN = 0.051
DOUBLE_CLICK_SECOND_HOLD_DELAY_STD  = 0.016
DOUBLE_CLICK_SECOND_HOLD_DELAY_KWARGS = {
    'loc': DOUBLE_CLICK_SECOND_HOLD_DELAY_MEAN,
    'scale': DOUBLE_CLICK_SECOND_HOLD_DELAY_STD 
}



class HLISAException(Exception):
  """ Base class of all exception classes used by HLISA.
  """
  pass
        
class OutOfViewportException(HLISAException):
  """ Indicates failing interaction due to leaving of the viewport
  """
  pass

class ElementBoundariesWereZeroException(HLISAException):
  """ Indicates that either the height or width of an element was zero
  """
  pass

class UnscrollableElementException(HLISAException):
  """ Indicates failing interaction due to leaving of the viewport
  """
  pass

class IllegalArgumentException(HLISAException):
  """ Indicates failing to provide legal arguments to a function
  """
  pass

class NoCursorCoordinatesException(HLISAException):
  """ An exception occured while retrieving the cursor location. 
      This should not happen, please file an issue if it did: https://github.com/droefs/HLISA/issues
  """
  pass



import math
import time
import random
import logging
import numpy as np

from selenium.webdriver import Firefox
from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys

from HLISA.selenium_actions import HL_Selenium_Actions
from HLISA.util import (behavorial_element_coordinates,
                        get_current_scrolling_position,
                        std_positive,
                        get_scrollable_elements,
                        element_is_scrollable)

from HLISA.errors import UnscrollableElementException

class HL_Additional_Actions:
    scroll_tick_size = 57

    def __init__(self, webdriver):
        self.webdriver = webdriver


    def shortPauze(self):
        """ Function that defines a short pause between actions
        """
        time.sleep(random.random() + 0.5)


    def move_to_element_outside_viewport(self, element, addDelayAfter=True):
        """ First scrolls to get the element into the viewport, then performs the movement
        """
        viewport_height = self.webdriver.execute_script("return window.innerHeight")
        y_relative = int(element.rect['y']) - get_current_scrolling_position(self.webdriver)["y"]
        y_diff = None
        if y_relative < 0 or y_relative > viewport_height:
            y_diff = y_relative if y_relative < 0 else y_relative - viewport_height/2

        if y_diff != None:
            self.scroll_by(0, y_diff)
            if get_current_scrolling_position(self.webdriver)["y"] == 0:
                # Fall back approach, if window.scrollBy does not work
                self.scroll_by_page_element(element, 0, y_diff)

        x, y = behavorial_element_coordinates(self.webdriver, element)
        selenium_actions = HL_Selenium_Actions(self.webdriver)
        selenium_actions.move_to(x, y, addDelayAfter)
        selenium_actions.perform()

    def scroll_by_page_element(self, element, x_diff, y_diff, addDelayAfter=True):
        """ Attempts scrolling via elements on the page

            @param element: the element of interest to search for scrollable parent
            elements
        """
        start = get_current_scrolling_position(self.webdriver, element)["y"]
        scrollable_elements = get_scrollable_elements(self.webdriver, element)
        while scrollable_elements and \
                start == get_current_scrolling_position(self.webdriver, element)["y"]:
            scroll = scrollable_elements.pop()
            self.scroll_by(x_diff, y_diff, addDelayAfter, scroll)


    def scroll_by(self, x_diff, y_diff, addDelayAfter=True, element=None):
        """ This function scrolls a few pixels further if the parameter is not a multiple of a standard scroll value.
            It would be detectable otherwise.
            Keyword arguments:
            x_diff -- the horizontal distance to scroll. 0 to not scroll horizontally.
            y_diff -- the vertical distance to scroll. 0 to not scroll vertically.
            element -- the element to scroll in (if None, the page is scrolled)
        """
        if x_diff != 0:
            raise NotImplementedError("Horizontal scrolling not yet implemented")
        if element is None or element_is_scrollable(self.webdriver, element):
            self.scroll_vertical(y_diff, element)
        else:
            raise UnscrollableElementException("The element can not be scrolled")
        if addDelayAfter:
            self.shortPauze()


    def scroll_vertical(self, y_diff_original, element=None):
        y_diff = y_diff_original
        scroll_ticks = 0
        current_y = get_current_scrolling_position(self.webdriver, element)["y"]
        if y_diff > 0:            
            while y_diff > 0:
                y_diff = self.scroll_tick(self.scroll_tick_size, scroll_ticks, y_diff, element)
                scroll_ticks += 1
        else:
            min_y = 0
            y_diff = max(y_diff, min_y - current_y) # Prevent scrolling too far
            while y_diff < 0:
                y_diff = self.scroll_tick((-1 * self.scroll_tick_size), scroll_ticks, y_diff, element)
                scroll_ticks += 1
        new_y = get_current_scrolling_position(self.webdriver, element)["y"]
        scrolled_distance = abs(current_y - new_y)
        if scrolled_distance == 0: # Scrolling is impossible
            return
        if scrolled_distance < abs(y_diff_original) - self.scroll_tick_size:
            if y_diff_original >= 0:
                self.scroll_vertical(y_diff_original - scrolled_distance)
            else:
                self.scroll_vertical(y_diff_original + scrolled_distance)


    def scroll_tick(self, pixelAmount, scroll_ticks, y_diff, element=None):
        """ Scrolls one tick on the window or a given element
        """
        if element:
            self.webdriver.execute_script("arguments[0].scrollBy(0, " + str(pixelAmount) + ")", element)
        else:
            self.webdriver.execute_script("window.scrollBy(0, " + str(pixelAmount) + ")")
        y_diff -= pixelAmount
        time.sleep(0.05 + (random.random()/200))
        if scroll_ticks % 7 == 0:
            time.sleep(std_positive(0.5, 0.1, 0))
        return y_diff


    def scroll_to(self, x, y, addDelayAfter=True):
        """ This function scrolls a few pixels further if the parameter is not a multiple of a standard scroll value.
            It would be detectable otherwise.
        """
        self.shortPauze()
        current_x = get_current_scrolling_position(self.webdriver)["x"]
        if current_x != x:
            raise NotImplementedError("Horizontal scrolling not yet implemented")
        current_y = get_current_scrolling_position(self.webdriver)["y"]
        y_diff = y - current_y
        self.scroll_by(x, y_diff, addDelayAfter)



import math
import time
import random
import numpy as np

from selenium.webdriver import Firefox
from selenium.webdriver.common.by import By
from selenium.webdriver.common.action_chains import ActionChains

from HLISA.selenium_actions import HL_Selenium_Actions
from HLISA.additional_actions import HL_Additional_Actions

class HLISA_ActionChains:
    """ This object holds its own chain of actions in self.chain.
        Every API call on this object adds the action to the chain,
        and if the action is ActionChain based itself (meaning it
        only executes after a .perform()), the .perform() is also
        called.
    """
    def __init__(self, webdriver, browser_resets_cursor_location=True):
        self.webdriver = webdriver
        self.chain = []

    ##### Standard Selenium action chain methods #####

    def click(self, element=None, addDelayAfter=True):
        actions = HL_Selenium_Actions(self.webdriver)
        self.chain.append(lambda: actions.click(element, addDelayAfter))
        self.chain.append(lambda: actions.perform())
        return self

    def click_and_hold(self, on_element=None, addDelayAfter=True):
        actions = HL_Selenium_Actions(self.webdriver)
        self.chain.append(lambda: actions.click_and_hold(on_element, addDelayAfter))
        self.chain.append(lambda: actions.perform())
        return self

    def context_click(self, on_element=None, addDelayAfter=True):
        actions = HL_Selenium_Actions(self.webdriver)
        self.chain.append(lambda: actions.context_click(on_element, addDelayAfter))
        self.chain.append(lambda: actions.perform())
        return self

    def double_click(self, on_element=None, addDelayAfter=True):
        actions = HL_Selenium_Actions(self.webdriver)
        self.chain.append(lambda: actions.double_click(on_element, addDelayAfter))
        self.chain.append(lambda: actions.perform())
        return self

    def drag_and_drop(self, source, target, addDelayAfter=True):
        actions = HL_Selenium_Actions(self.webdriver)
        self.chain.append(lambda: actions.drag_and_drop(source, target, addDelayAfter))
        self.chain.append(lambda: actions.perform())
        return self

    def drag_and_drop_by_offset(self, source, xoffset, yoffset, addDelayAfter=True):
        actions = HL_Selenium_Actions(self.webdriver)
        self.chain.append(lambda: actions.drag_and_drop_by_offset(source, xoffset, yoffset, addDelayAfter))
        self.chain.append(lambda: actions.perform())
        return self

    def key_down(self, value, element=None, addDelayAfter=True):
        actions = HL_Selenium_Actions(self.webdriver)
        self.chain.append(lambda: actions.key_down(value, element, addDelayAfter))
        self.chain.append(lambda: actions.perform())
        return self

    def key_up(self, value, element=None, addDelayAfter=True):
        actions = HL_Selenium_Actions(self.webdriver)
        self.chain.append(lambda: actions.key_up(value, element, addDelayAfter))
        self.chain.append(lambda: actions.perform())
        return self

    def move_by_offset(self, x, y, addDelayAfter=True):
        actions = HL_Selenium_Actions(self.webdriver)
        self.chain.append(lambda: actions.move_by_offset(x, y, addDelayAfter))
        self.chain.append(lambda: actions.perform())
        return self

    def move_to_element(self, element, addDelayAfter=True):
        actions = HL_Selenium_Actions(self.webdriver)
        self.chain.append(lambda: actions.move_to_element(element, addDelayAfter))
        self.chain.append(lambda: actions.perform())
        return self

    def move_to_element_with_offset(self, to_element, xoffset, yoffset, addDelayAfter=True):
        actions = HL_Selenium_Actions(self.webdriver)
        self.chain.append(lambda: actions.move_to_element_with_offset(to_element, xoffset, yoffset, addDelayAfter))
        self.chain.append(lambda: actions.perform())
        return self

    def pause(self, seconds):
        actions = HL_Selenium_Actions(self.webdriver)
        self.chain.append(lambda: actions.pause(seconds))
        self.chain.append(lambda: actions.perform())
        return self

    def perform(self):
        for action in self.chain:
            action()
        if HL_Selenium_Actions.selenium_version >= 4:
            self.reset_actions()
        return self

    def release(self, on_element=None, addDelayAfter=True):
        actions = HL_Selenium_Actions(self.webdriver)
        self.chain.append(lambda: actions.release(on_element, addDelayAfter))
        self.chain.append(lambda: actions.perform())
        return self

    def reset_actions(self):
        self.chain = []
        return self

    def send_keys(self, keys_to_send, element=None, addDelayAfter=True, speed_scaling=1.0):
        actions = HL_Selenium_Actions(self.webdriver)
        self.chain.append(lambda: actions.send_keys(keys_to_send, element, addDelayAfter, speed_scaling=speed_scaling))
        self.chain.append(lambda: actions.perform())
        return self

    def send_keys_to_element(self, element, keys_to_send, addDelayAfter=True, speed_scaling=1.0):
        actions = HL_Selenium_Actions(self.webdriver)
        self.chain.append(lambda: actions.send_keys_to_element(element, keys_to_send, addDelayAfter, speed_scaling=speed_scaling))
        self.chain.append(lambda: actions.perform())
        return self

    ##### Additional actions #####

    def move_to(self, x, y, addDelayAfter=True):
        actions = HL_Selenium_Actions(self.webdriver)
        self.chain.append(lambda: actions.move_to(x, y, addDelayAfter))
        self.chain.append(lambda: actions.perform())
        return self

    def move_to_element_outside_viewport(self, element, addDelayAfter=True):
        additional_actions = HL_Additional_Actions(self.webdriver)
        self.chain.append(lambda: additional_actions.move_to_element_outside_viewport(element, addDelayAfter))
        return self

    def scroll_by(self, x_diff, y_diff, addDelayAfter=True, element=None):
        additional_actions = HL_Additional_Actions(self.webdriver)
        self.chain.append(lambda: additional_actions.scroll_by(x_diff, y_diff, addDelayAfter, element))
        return self

    def scroll_to(self, x_diff, y_diff, addDelayAfter=True):
        additional_actions = HL_Additional_Actions(self.webdriver)
        self.chain.append(lambda: additional_actions.scroll_to(x_diff, y_diff, addDelayAfter))
        return self

    def back(self):
        self.webdriver.back()
        return self

    def forward(self):
        self.webdriver.forward()
        return self
